package GraphEdit::GraphNode ;

use strict ;
use warnings ;
use Carp ;
use Data::Dumper ;
use GraphEdit::Utils qw/ gv_pos_to_coord / ;

use gv ;

my $id = 1 ;
my ( $prevx, $prevy ) ; # used when moving one or more nodes
my ( $startx, $starty ) ;

# ----------------------------------------------------------------------
#                         class constructor
# ----------------------------------------------------------------------

sub new {

  my $class = shift ;
  my %pars  = @_ ;

  my $pos   = $pars{pos} || [ 20, 20 ] ;

  unless( $pars{canvas} )   { die "canvas not defined" ; }
  unless( $pars{mw} )       { die "mw not defined" ; }
  unless( $pars{domgraph} ) { die "domgraph not defined" ; }

  my $canvas ||= $pars{canvas} ;

  my $self = bless {
    id          => $pars{id} || $id,
    mw          => $pars{mw},
    canvas      => $pars{canvas},
    nodemaster  => $pars{nodemaster},
    domgraph    => $pars{domgraph},
    label       => $pars{label} || '?',
    x           => $pos->[0],
    y           => $pos->[1],
    links       => { },
    link_no     => 0,
    gvnode      => $pars{gvnode}  || undef,
    color       => $pars{color}   || '#eeeeee',
    fgcolor     => $pars{fgcolor} || '#000000',
    meta_info   => $pars{meta_info} || '',
  } ;

  #print "node object id $pars{id} label $title tag dnode_$id canvas $poly\n" ;

  $self->node_create_graphics( $pos->[0], $pos->[1], $pars{width}, $pars{height} ) ;

  #carp "Created new dnode object id $id" if ( $::DEBUG ) ;
  $id++ ;

  return $self ;
}






# ----------------------------------------------------------------------
#                         various functions
# ----------------------------------------------------------------------

sub node_create_graphics {
  my $self = shift ;
  my ( $posx, $posy, $width, $height ) = @_ ;

  my $id     = $self->{id} ;
  my $canvas = $self->{canvas} ;

  my $text_id = $canvas->createText( $posx,
    $posy,
    -anchor => 'center', 
    -text => $self->{label}, 
    -fill => $self->{fgcolor},
    -tags => [ 'dnode', "dnode_$id", "tdnode_$id" ],) ;

  my @bb = $canvas->bbox( "tdnode_$id" ) ;

  my $w = $width  || ( $bb[2] - $bb[0]  + 20 ) ;
  my $h = $height || ( $bb[3] - $bb[1]  + 20 ) ;

  @bb = ( $posx - $w / 2, $posy - $h / 2, 
          $posx + $w / 2, $posy - $h / 2,
          $posx + $w / 2, $posy + $h / 2,
          $posx - $w / 2, $posy + $h / 2, ) ;

  my $poly = $self->{canvas}->createRectangle(
    $posx - $w / 2, $posy - $h / 2,
    $posx + $w / 2, $posy + $h / 2,

    #@bb,
    -fill    => $self->{color},
    -tags    => [ 'tyPol', 'dnode', "dnode_$id" ],
    -outline => 'black',
  );

  $self->{cv_id}   = "dnode_$id" ;
  $self->{text_id} = $text_id ;
  $self->{box_id}  = $poly ;

  $canvas->raise( "dnode_$id", 'all' ) ;
  $canvas->raise( "tdnode_$id", "dnode_$id" ) ;

}

# return the number of links of a node
sub nlinks {
  return $_[0]->{link_no} ;
}

# check whether the node and partner interact or modify interaction strength
sub islink {
  my ( $self, $partner, $s ) = @_ ;

  return 0 unless $self->{links}{$partner} ;

  if( defined $s ) {
    $self->{links}{$partner}[2] = $s ;
    $partner->{links}{$self}[2] = $s ;
    my $id = $self->{links}{$partner}[0] ;
    #print "new width for $id: $s\n" ;
    $self->{canvas}->itemconfigure( $id, '-width', $s ) ;
    $self->{canvas}->itemconfigure( $id, '-fill', "black" ) ;
  } 

  return $self->{links}{$partner}[2] || 1 ; 
}

# create a new link, graphics plus entry in the partners structure
sub addlink {
  my ( $self, $partner, %pars ) = @_ ;
  #print "adding link $self - $partner\n" if( $::DEBUG ) ;
  return if $self == $partner ;

  my $cur_s = $self->islink( $partner ) ;
  if( $cur_s ) { return $cur_s ; }

  $self->{link_no}++ ;
  $partner->{link_no}++ ;

  #print "* creating new link $self - $partner\n" ;
  my $id = $self->draw_link( $partner, %pars ) ;
  $self->{links}{$partner} = [ $id, $partner ] ;
  $partner->{links}{$self} = [ $id, $self ] ;

}

sub draw_link {
  my $self    = shift ;
  my $partner = shift ;
  my %pars    = @_ ;

  my @coords = ( $self->x, $self->y, $partner->x, $partner->y ) ;
  #print STDERR "drawing link at coords (".join(',', @coords).") pars" .  join(' ', %pars ) ."\n" ;
  my $id = $self->{canvas}->createLine( @coords, %pars ) ;
  $self->{canvas}->lower( $id, 'all' ) ;

  return $id ;
}





# ----------------------------------------------------------------------
# a bunch of methods for setting / querying state of the object
# ----------------------------------------------------------------------

sub x { # return the x coordinate
  return $_[0]->{x} ;
}

sub y { # return the y position
  return $_[0]->{y} ;
}

sub gvnode {
  return $_[0]->{gvnode} ;
}






# ----------------------------------------------------------------------
#                        moving node graphics
# ----------------------------------------------------------------------
# scale the position of the node by factor $fac respective to the origin
# (xorig,yorig)

sub scale_pos {
  my $self = shift ;

  my ( $fac, $xorig, $yorig)  = @_ ;
  $xorig ||= 0 ; $yorig ||= 0 ;

  my ( $x, $y ) = gv_pos_to_coord( gv::getv( $self->{gvnode}, 'pos' ) ) ;
  my ( $x1, $y1 ) = ( $self->{x}, $self->{y} ) ;
  $y = -$y ;

  $x -= $xorig ;
  $y -= $yorig ;

  my ( $dx, $dy ) = ( $fac * $x - $x, $fac * $y - $y) ;
  $self->move( $dx, $dy ) ;

}

sub _move_graphics_rel {
  my $self = shift ;
  my ( $dx, $dy ) = @_ ;

  my $id = $self->{cv_id} ;

  $self->{canvas}->move( $id, $dx, $dy ) ;
  $self->{x} += $dx ;
  $self->{y} += $dy ;

  my @c = ( $self->x, $self->y ) ;
  foreach my $link ( values %{ $self->{links} } ) {
    my ( $id, $l ) = @$link ;
    $c[2] = $l->x ;
    $c[3] = $l->y ;
    $self->{canvas}->coords( $id, @c ) ;
  }

}

# move one point of $edge by [$dx,$dy]
# which point? the one which is closer to $node_pos (which is a string
# returned by gv:getv( $node, 'pos' ) 
sub _move_gv_edge_rel {
  my $self = shift ;
  my ( $edge, $dx, $dy, $node_pos ) = @_ ;

  #print "node_pos: $node_pos\n" ;
  my ( $nx, $ny ) = gv_pos_to_coord( $node_pos ) ;
  #print "nx=$nx ny=$ny\n" ;

  my $epos = gv::getv( $edge, 'pos' ) ;
  my @epos_l = split /\s+/, $epos ;

  my $min_e ; my $min_ed ; my $max_e ; my $max_ed ;
  foreach my $e ( @epos_l ) {
    unless( $e ) {
      die "blah! >$e< $epos\n" ;
    }
    my ( $ex, $ey ) = gv_pos_to_coord( $e ) ;
    unless( defined $ex and defined $ey ) {
      die "blah! $e|$epos\n" ;
    }


    my $d = sqrt( ( $ex - $nx ) ** 2 + ( $ey - $ny ) ** 2 ) ;
    #print "$e $ex $ey d=$d\n" ;
    if( not defined $max_ed or $max_ed < $d ) {
      $max_ed = $d ;
      $max_e  = [ $ex, $ey ] ;
    }

    if( not defined $min_ed or $min_ed > $d ) {
      $min_ed = $d ;
      $min_e  = [ $ex, $ey ] ;
    }


  }
  #print "max=$max_e->[0] $max_e->[1] min=$min_e->[0] $min_e->[1]\n" ;

  $min_e->[0] += $dx ;
  $min_e->[1] += $dy ;

  my $ret = gv::setv( $edge, 'pos', "$min_e->[0],$min_e->[1] $max_e->[0],$max_e->[1]" ) ;
  #print "ret=$ret\n" ;
}

sub _move_gv_rel {
  my $self = shift @_  ;
  my ( $dx, $dy ) = @_ ;

  my $pos = gv::getv( $self->{gvnode}, 'pos' ) ;
  my ( $x, $y ) = gv_pos_to_coord( $pos ) ;
  #print "moving $self->{label} from $pos = ($x,$y) " ;
  $x -= $dx ; $y += $dy ;

  my $newpos = sprintf( "%d,%d", $x, $y ) ;
  #print "to ($x,$y) = $newpos\n" ;
  gv::setv( $self->{gvnode}, 'pos', $newpos ) ;
  gv::setv( $self->{gvnode}, 'pin', 'true' ) ;

  my $n = $self->{gvnode} ;

  for( my $edge = gv::firstedge( $n ) ; gv::ok( $edge ) ; $edge = gv::nextedge( $n, $edge ) ) {

    $self->_move_gv_edge_rel( $edge, $dx, $dy, $pos ) ;

  }
}

sub _move_gv_abs {
  my $self = shift ;
  my ( $x, $y ) = @_ ;

  my $pos = gv::getv( $self->{gvnode}, 'pos' ) ;
  my ( $sx, $sy ) = gv_pos_to_coord( $pos ) ;

  my $newpos = sprintf( "%f,%f", $x, -$y ) ;
  my $gvnode = $self->{gvnode} ;

  gv::setv( $gvnode, 'pos', $newpos ) ;
  gv::setv( $gvnode, 'pin', 'true' ) ;

  for( my $edge = gv::firstedge( $gvnode ) ; gv::ok( $edge ) ; $edge = gv::nextedge( $gvnode, $edge ) ) {
    $self->_move_gv_edge_rel( $edge, $x - $sx, $y - $sy, $pos ) ;
  }

}

# move the dnode object to a new position on the canvas
sub move {
  my $self = shift @_  ;
  my ( $dx, $dy ) = @_ ;

  $self->_move_graphics_rel( $dx, $dy ) ;

  if( $self->{gvnode} ) {
    $self->_move_gv_rel( -$dx, -$dy ) ;
  } 
}

sub update_pos_from_graph {
  my $self = shift ;

  return unless $self->{gvnode} ;
  my $id     = $self->{id} ;
  my $canvas = $self->{canvas} ;

  my $pos = gv::getv( $self->{gvnode}, 'pos' ) ;
  my ( $x, $y ) = gv_pos_to_coord( $pos ) ;
  $y = -$y ;

  my @bb = $canvas->bbox( $self->{cv_id} ) ;
  my ( $cvx, $cvy ) = (  ( $bb[0] + $bb[2] ) / 2, ( $bb[1] + $bb[3] ) / 2 ) ;
  my ( $dx, $dy ) = ( $x - $cvx, $y - $cvy ) ;

  $self->_move_graphics_rel( $dx, $dy ) ;
}






# ----------------------------------------------------------------------
# var node operations
# ----------------------------------------------------------------------

sub links {
  my $self = shift ;
  return $self->{links} ;
}

sub _delete {
  my $self = shift ;

  print STDERR "Deleting node $self->{label}\n" ;
  $self->unmark() ;

  my @lofparts = keys %{ $self->{links} } ;
  foreach my $partner ( @lofparts ) {
    print "deleting link to partner $partner\n" ;

    my ( $id, $part, $strength ) = @{ $self->{links}{$partner} } ;
    $self->{canvas}->delete( $id ) ;
    print "my links: " . $self->links() . "\n" ;
    print "me: $self\n" ;
    my $partlinks = $part->links() ;

    print "in partner: " . join( ' ', keys %$partlinks ) .  "\n" ;
    delete $part->{links}{$self} ;
    print "in partner now: " . join( ' ', keys %$partlinks ) .  "\n" ;

  }

  $self->{domgraph}->network_changed() ;

  $self->{canvas}->delete( $self->{cv_id} ) ;

  if( $self->{gvnode} ) { gv::rm( $self->{gvnode} ) ; }

}






# ----------------------------------------------------------------------
# handling selections
# ----------------------------------------------------------------------


# remove the marking from a node
sub unmark {
  my $self = shift ;
  my $mark = $self->{mark} ;
  return unless $mark ;

  $self->{canvas}->delete( $mark ) ;
  $self->{mark} = 0 ;
}

# put a yellow marking around a node
sub mark {
  my $self = shift ;
  return if( $self->{mark} ) ;

  my $tag = $self->{cv_id} ;
  my $canvas = $self->{canvas} ;

  my @bb  = $canvas->bbox( $tag ) ;
  unless( @bb ) {
    print "WARNING. No bbox found for tag $tag\n" ;
    return ;
  }
  my $m = 2 ;
  @bb = ( $bb[0] - $m, $bb[1] - $m, $bb[2] + $m, $bb[3] + $m ) ;
  #@bb = ( $bb[0], $bb[1], $bb[2], $bb[1], $bb[2], $bb[3], $bb[0], $bb[3] ) ;

  my $mark = $canvas->createRectangle(
    @bb,
    -fill    => 'yellow',
    -tags    => [ 'tyPol', 'mark', $tag ],
    -outline => '',
  );
  $canvas->lower( $mark, 'all' ) ;
  $self->{mark} = $mark ;

}


# ----------------------------------------------------------------------
# configuring various graphical parameters
# ----------------------------------------------------------------------


# create a link between a node and another node_id
sub link_configure {
  my $self        = shift ;
  my $partnode_id = shift ;
  my %pars = @_ ;
  my $partnode = $self->{nodemaster}->{node_hash}->{ $partnode_id } ;

  my $link = $self->{links}{$partnode} ;
  $self->{canvas}->itemconfigure( $link->[0], %pars ) ;
}

# configure the properties of a node
sub configure {
  my $self = shift ;
  my %pars = @_ ;

  while( my ( $k, $v ) = each %pars ) {
    #print "k=$k, v=$v\n" ;
    next unless $v ;
    if( $k =~ /^(color|fill)$/ ) {
      $self->set_color( $v ) ;
    } elsif( $k eq 'label' ) {
      $self->set_label( $v ) ;
    } else {
      $self->{$k} = $v ;
      print "setting $k to $v\n" ;
    }
  }

}

# change the label of a node
sub set_label {
  my $self = shift ;
  my $lab  = shift ;

  my $canvas = $self->{canvas} ;

  $self->{label} = $lab ;
  $canvas->itemconfigure( $self->{text_id}, -text => $lab ) ;

  # update box size
  my @bb = $canvas->bbox( $self->{text_id} ) ;
  $canvas->coords( $self->{box_id}, @bb ) ;

  if( my $gvnode = $self->{gvnode} ) { gv::setv( $gvnode, 'label', "$lab" ) ; }
}

# change the color of a node
sub set_color {
  my $self = shift ;
  my $col  = shift ;

  $self->{canvas}->itemconfigure( $self->{box_id}, -fill => $col ) ;
  $self->{color} = $col ;
  if( my $gvnode = $self->{gvnode} ) { gv::setv( $gvnode, 'color', "$col" ) ; }
}

# report back the color of a node
sub color {
  my $self = shift ;
  #my $col = $self->{canvas}->itemcget( $self->{box_id}, -fill ) ;
  return $self->{color} ;
}




1; 
