package GraphEdit::GraphWindow ;
use strict   ;
use warnings ;

use gv ;
use Data::Dumper ;

use Tk              ;
use Tk::Canvas      ;
use Tk::Pod         ;

#use Tk::Balloon     ;
#use Tk::BrowseEntry ;
#use Tk::Clipboard   ;
#use Tk::Dialog      ;
#use Tk::FileSelect  ;
#use Tk::ProgressBar ;
#use Tk::X11Font     ;

use GraphEdit::Menufactory qw/ create_menu / ;
use GraphEdit::Nodemaster ;

# only in development:
#use Devel::DProf ;

my $DEBUG = 1 ;
my $networkchanged = 0 ;

# ----------------------------------------------------------------------
#                        Grahphviz defaults
# ----------------------------------------------------------------------

my $gv_defaults = {
  N => {
    shape => 'box',
    style => 'filled',
    margin => 0.01,
    fontsize => 8,
    width => 0,
    height => 0, },
  G => { 
    outputorder => 'edgesfirst',
    },
  E => { },
} ;



# ----------------------------------------------------------------------
#                        CLASS constructor
# ----------------------------------------------------------------------

sub new {
  my $class = shift ;
  my %pars  = @_ ;
  my $mw = $pars{-mw} or die "-mw parameter missing" ;

  my $self = bless {
    mw          => $mw,
    canvas      => '',
    minlinks    => 2,
    layout      => 'nop2',
    gv_defaults => $gv_defaults,
  } ;

  $self->_create_main_window( %pars ) ;

  my $Y = $self->{canvas}->cget( '-height' ) ;
  my $X = $self->{canvas}->cget( '-width' ) ;

  $self->{width}  = $X ;
  $self->{height} = $Y ;

  $self->{mw}->update ;
  $self->{canvas}->update ;
  if( $pars{-graph} ) {
    $self->new_graph( $pars{-graph} ) ;
  } else {
    $self->{canvas}->configure( -bg => '#eeeeee' ) ;
  }

  $self->{mw}->update();
  return $self ;
}

# ----------------------------------------------------------------------
#                        Reading graph files
# ----------------------------------------------------------------------

sub new_graph {
  my $self  = shift ;
  my $graph = shift ;

  my $mw = $self->{mw} ;
  $self->{canvas}->configure( -bg => 'white' ) ;
  $self->{nodemaster} = GraphEdit::Nodemaster->new( mw => $mw, canvas => $self->{canvas}, domgraph => $self  ) ;
  $self->{graph} = $self->gv_graph_new() ; #$pars{-graph} || '' ;
  $self->copy_graph( $graph, $self->{graph} ) ;
  $self->_create_graph( ) ;
  $self->reconfigure_canvas() ;
}

sub delete_current {
  my $self = shift ;
  $self->{canvas}->delete( 'all' ) ;
}

sub open_dot {
  my $self = shift ;

  my $types = [ [ 'Dot files', [ '.dot' ] ], [ 'All files', '*' ] ] ;
  my $fname = $self->{mw}->getOpenFile( -filetypes => $types ) ;
  return unless( $fname ) ;

  print STDERR "Reading graph from file $fname\n" ;

  if( $self->{graph} ) {
    $self->delete_current() ;
  }

  my $graph  = gv::read( $fname ) ;
  unless( $graph ) {
    error_box( $self->{mw}, "Could not read graph from file $fname" ) ;
    return ;
  }

  $self->new_graph( $graph ) ;


}



# ----------------------------------------------------------------------
#                        Grahphviz operations
# ----------------------------------------------------------------------

# create a new graph
sub gv_graph_new {
  my $self = shift ;

  my $graph = gv::graph( 'Unnamed' ) ;
  my $N     = gv::protonode( $graph ) ;
  my $E     = gv::protoedge( $graph ) ;

  my $params = $self->{gv_defaults} ;

  # setting default parameters
  while( my ( $k, $v ) = each %{ $params->{G} } ) { gv::setv( $graph, $k, $v ) ; }
  while( my ( $k, $v ) = each %{ $params->{N} } ) { 
    print "setting: $k-$v\n"  ; gv::setv( $N, $k, $v ) ; }
  while( my ( $k, $v ) = each %{ $params->{E} } ) { gv::setv( $E, $k, $v ) ; }

  return $graph ;
}

# copy a whole graph structure into a new graph
sub copy_graph {
  my $self = shift ;
  my ( $src, $dest ) = @_ ;

  # copy graph attributes
  gv_copy_graph_attr( $src, $dest ) ;

  for( my $snode = gv::firstnode( $src ) ; gv::ok( $snode ) ; $snode = gv::nextnode( $src, $snode ) ) {
    gv_copy_node( $src, $dest, $snode ) ;
  }

  my %seen ;
  for( my $snode = gv::firstnode( $src ) ; gv::ok( $snode ) ; $snode = gv::nextnode( $src, $snode ) ) {
  for( my $sedge = gv::firstedge( $snode ) ; gv::ok( $sedge ) ; $sedge = gv::nextedge( $snode, $sedge ) ) {

    my $nodet = gv::tailof( $sedge ) ;
    my $nodetname = gv::nameof( $nodet ) ;
    my $nodeh = gv::headof( $sedge ) ;
    my $nodehname = gv::nameof( $nodeh ) ;
    next if( $seen{ "$nodetname-$nodehname" } ) ;
    $seen{ "$nodetname-$nodehname" }++ ;
    gv_copy_edge( $src, $dest, $sedge, $nodehname, $nodetname ) ;
  }}

}

# set all nodes to a specified value
sub _gv_set_node_global {
  my ( $graph, $attrname, $value ) = @_ ;

  for( my $node = gv::firstnode( $graph ) ; gv::ok( $node ) ; $node = gv::nextnode( $graph, $node ) ) {
    gv::setv( $node, $attrname, $value ) ;
  }

}

# copy an edge with all its attributes
sub gv_copy_edge {
  my ( $src, $dest, $sedge, $nodehname, $nodetname ) = @_ ;
  my $dedge = gv::edge( $dest, $nodehname, $nodetname ) ;

  for( my $a = gv::firstattr( $sedge ) ; gv::ok( $a ) ; $a = gv::nextattr( $sedge, $a ) ) {
    my $v = gv::getv( $sedge, $a ) ;
    my $aname = gv::nameof( $a ) ;
    #print STDERR "copying $aname = $v\n" ;
    gv::setv( $dedge, $aname, $v ) ;
  }
  return $dedge ;
}

# copy attributes of a graph
sub gv_copy_graph_attr {
  my ( $src, $dest ) = @_ ;

  for( my $a = gv::firstattr( $src ) ; gv::ok( $a ) ; $a = gv::nextattr( $src, $a ) ) {
    my $v = gv::getv( $src, $a ) ;
    my $aname = gv::nameof( $a ) ;
    gv::setv( $dest, $aname, $v ) ;
  }

}

# copy a node and all its attributes
sub gv_copy_node {
  my ( $src, $dest, $snode ) = @_ ;

  my $snodenam = gv::nameof( $snode ) ;
  #print STDERR "Copying node $snodenam\n" ;
  my $dnode    = gv::node( $dest, $snodenam ) ;

  for( my $a = gv::firstattr( $snode ) ; gv::ok( $a ) ; $a = gv::nextattr( $snode, $a ) ) {
    my $v = gv::getv( $snode, $a ) ;
    my $aname = gv::nameof( $a ) ;
    #print STDERR "copying $aname = $v\n" ;
    gv::setv( $dnode, $aname, $v ) ;
  }


  return $dnode ;
}





# ----------------------------------------------------------------------
#                        CREATE THE MAIN GRAPH
# ----------------------------------------------------------------------

# create the node objects on the canvas
sub _create_nodes {
  my $self      = shift ;

  my $canvas    = $self->{canvas} ;
  my $graph     = $self->{graph}  ;
  my $w         = $self->{width}  ;
  my $h         = $self->{height} ;

  return unless $graph ;

  my %nodeobjects ;
  my ( $links, $nl ) ;
  print STDERR "Creating nodes\n" ;
  my $nid = 1 ;

  for( my $curnode = gv::firstnode( $graph ) ; gv::ok( $curnode ) ; $curnode = gv::nextnode( $graph, $curnode ) ) {

    my $nodeid = gv::nameof( $curnode ) ;
    my $label  = $nodeid ;
    my $tmp ;
    if( $tmp = gv::getv( $curnode, 'label' ) and $tmp ne '\N' ) { $label = $tmp ; }
    else{ gv::setv( $curnode, 'label', $label ) or die "Cannot set attribute label" ; } ;

    my ( $x, $y ) = ( int( rand( $w )), int( rand( $h ) ) ) ;
    if( my $pos = gv::getv( $curnode, 'pos' ) ) {
      ( $x, $y ) = ( $pos =~ /(\S+),(\S+)/ ) ;
      $y = -$y ;
      unless( $x and $y ) { die "Cannot parse attr value: $pos" ; }
    } 

    my $dpi   = gv::getv( $graph, 'dpi' ) || 96 ;
    my $nw    = gv::getv( $curnode, 'width' ) * $dpi ;
    my $nh    = gv::getv( $curnode, 'height' ) * $dpi ;
    my $color = gv::getv( $curnode, 'color' ) || '' ;

    $nodeobjects{$nodeid} = 
        $self->{nodemaster}->new_node(  
          id       => $nodeid,
          mw       => $self->{mw},
          domgraph => $self,
          canvas   => $canvas, 
          color    => $color,
          gvnode   => $curnode,
          width    => $nw,
          height   => $nh,
          pos      => [ $x, $y ],
          label    => $label ) ;

    $nid++ ;
  }

  print STDERR "Done creating nodes\n" ;
  return \%nodeobjects ;
}

# creates links between the nodes
sub _create_links {
  my $self = shift ;
  #my ( $nodeobjects ) = @_ ;
  my $nodeobjects = $self->{nodemaster}->node_hash ;
  return unless( $self->{graph} ) ;
  my $graph = $self->{graph} ;
  #print STDERR "Creating links...\n" ;

  while( my ( $nodeid, $node ) = each %$nodeobjects ) {
    #print STDERR "nodeid $nodeid\n" ;
    my $gvnode = $node->gvnode ;
    next unless( $gvnode ) ;

    #print "node: $label\n" ;
    for( my $edge = gv::firstedge( $gvnode ) ; gv::ok( $edge ) ; $edge = gv::nextedge( $gvnode, $edge ) ) {

      #print "edge\n" ;
      my $partnodeid ;
      my $hnodeid = gv::nameof( gv::headof( $edge ) ) ;
      my $tnodeid = gv::nameof( gv::tailof( $edge ) ) ;
      if( $hnodeid eq $nodeid ) {
        $partnodeid = $tnodeid ;
      } else {
        $partnodeid = $hnodeid ;
      }

      #print "node $nodeid part $partnodeid\n" ;
      $node->addlink( $nodeobjects->{$partnodeid} ) ;
      next ;

    }
  }

  print STDERR "Done creating links\n" ;
  return ;

}

# create the full graph
sub _create_graph {
  my $self = shift ;

  return unless $self->{graph} ;

  $self->_render_graph() ;

  my ( $nodehash ) = $self->_create_nodes( ) ;
  $self->{nodehash} = $nodehash ;
  $self->_create_links( $nodehash ) ;

  my $nodemaster = $self->{nodemaster} ;

  # make sure that clicking of the canvas doesn't cause the selection
  # rectangle to appear
  $self->{canvas}->bind( 'all', '<ButtonPress>',   sub { $nodemaster->{item_clicked} = 1 }  ) ;
  $self->{canvas}->bind( 'all', '<ButtonRelease>', sub { $nodemaster->{item_clicked} = 0 }  ) ;

}

# create the layout of the graph, render
sub _render_graph {
  my $self = shift ;

  print STDERR "Rendering...\n" ;

  my $error = 0 ;
  my $graph = $self->{graph} ;
  unless( $graph ) {
    die "graph undefined\n" ;
  }

  for( my $node = gv::firstnode( $graph ) ; gv::ok( $node ) ; $node = gv::nextnode( $graph, $node ) ) {
    my $pos = gv::getv( $node, 'pos' ) ;
    unless( $pos ) { $error++ ; last ; }
  }

  if( $error and ( $self->{layout} eq 'nop' or $self->{layout} eq 'nop2' )) {
    print STDERR "Positions undefined in graph: using neato layout\n" ; 
    $self->{layout} = 'neato' ;
  }

  gv::layout( $graph, $self->{layout} || 'dot' ) ;
  gv::render( $graph ) ;
  print STDERR "done\n" ;


}

# ----------------------------------------------------------------------

sub new_layout {
  my $self   = shift ;
  my $layout = shift ;

  return unless( $self->{graph} ) ;
  my $graph = $self->{graph} ;

  my $c = $self->{mw}->cget( '-cursor' ) ;
  $self->{mw}->configure( -cursor => 'watch' ) ;
  $self->{mw}->update ;
  print STDERR "Rendering graph in layout $layout\n" ;
  gv::layout( $graph, $layout ) ;
  gv::render( $graph ) ;
  print STDERR "Done\n" ;

  foreach my $n ( values %{ $self->{nodehash} } ) {
    $n->update_pos_from_graph() ;
  }
  $self->{mw}->configure( -cursor => $c ) ;

  $self->zoom_auto() ;


}





# ----------------------------------------------------------------------
# GUI: perl-tk stuff to create the canvased window
# ----------------------------------------------------------------------

sub _create_main_window {
  my $self = shift ;
  my %pars = @_;

  my $mw     = $self->{mw} ; #->Toplevel( -title => 'Domain graph' );

  my $scrollreg = $mw->Scrolled(
    'Canvas',
    -background   => 'white',
    -scrollregion => [ 0, 0, 600, 400 ],
    -width        => 600,
    -height       => 300,
  );

  my $canvas         = $scrollreg->Subwidget('scrolled');
  $self->{scrollreg} = $scrollreg ;
  $self->{canvas}    = $canvas ;
  $self->{mw}        = $mw ;

  my $menus = $self->window_menu( %pars, -mw => $mw );

  $self->{menu_widgets} = 
    create_menu( %pars, -mw => $mw, -menus => $menus ) ;

  $scrollreg->pack(   -expand => 1, -fill => 'both' ) ;
  $mw->update();

}

# just the menu setup
sub window_menu {
  my $self = shift ;
  my $mw   = $self->{mw} ;

  # for each menu item, we have ref. to array with the following items:
  # 0 - function bind to the item; 1 - global shortcut ; 2 - optional type
  # 3 - optional variable for the radiobutton
  my @menus = (

    'File' => [
      '~Open DOT file'          => [ sub { $self->open_dot() }, 'Ctrl o', ],
      '-'                       => [ '', '', 'separator' ],
      '~Preview'                => [ sub { $self->preview_graph() }, 'Ctrl p', ],
      'Export PDF'              => [ sub { $self->export( 'PDF' ) } ],
      'Export PNG'              => [ sub { $self->export( 'PNG' ) } ],
      '-'                       => [ '', '', 'separator' ],
      '~Save DOT graph'         => [ sub { $self->save() }, 'Ctrl s', ],
      'Save DOT graph ~as'      => [ sub { $self->save( 1 ) }, 'Ctrl S', ],
      '-'                       => [ '', '', 'separator' ],
      "~Quit"                   => [ sub { $mw->destroy }, 'Ctrl q' ],
    ], 

    'Import' => [
       '~Meta info'     => [ sub { $self->i_read_generic( fields => [ 'meta_info' ] ) },  ],
       '~Colors'        => [ sub { $self->i_read_generic( fields => [ 'color' ] ) }, ],
       '~Labels'        => [ sub { $self->i_read_generic( fields => [ 'label'] ) }, ],
       '~Automatic (with header)'     
                        => [ sub { $self->i_read_generic( ) }, ],
       '~Add graph'     => [ sub { $self->graph_compare( add => 1 ) } ],
    ],

    'Graph' => [
       '~Select All'                         => [ sub { $self->select_all() }, 'Ctrl a', ],
       '~Find'                               => [ sub { $self->find_dialog() }, 'Ctrl f', ],
       "~Remove small clusters"              => [ sub { $self->_remove_small_clusters( 2 ) }, ],
       "~New graph from selection"           => [ sub { $self->_new_graph_from_selection() } ],
       "C~ompare with another graph"         => [ sub { $self->graph_compare() } ],
    ],

    Layout => [
      '~neato' => [ sub { $self->new_layout( 'neato' ) } ],
      '~dot'   => [ sub { $self->new_layout( 'dot' ) } ],
      '~twopi' => [ sub { $self->new_layout( 'twopi' ) } ],
      '~circo' => [ sub { $self->new_layout( 'circo' ) } ],
      '~fdp'   => [ sub { $self->new_layout( 'fdp' ) } ],
    ],

    'Display' => [
      'Zoom ~in'   => [ sub { $self->zoom_in()   }, 'Ctrl +',  ],
      'Zoom ~out'  => [ sub { $self->zoom_out()  }, 'Ctrl -',  ],
      'Zoom ~auto' => [ sub { $self->zoom_auto() }, 'Ctrl z',  ],
      '~Redraw'    => [ sub { $self->reconfigure_canvas() }, 'Ctrl L',  ],
    ],

    'Help' => [
      "~Instructions" => [ sub { $self->view_help() }, 'F1' ],
      "~About"  => [ "", ],
    ],

  );

  return \@menus;
}

# reconfigure canvas: (i) resize the scrollregion and (ii) bind mouse/keys
sub reconfigure_canvas {
  my $self = shift ;
  my $canvas = $self->{canvas} ;
  my %pars = @_ ;
  
  my ( $sizex, $sizey ) = ( $canvas->bbox('all') )[ 2, 3 ];
  my @bbox = ( $canvas->bbox('all') ) ;
  print STDERR "size is now $sizex x $sizey\n" if ($::DEBUG);
  print STDERR "reconfigure_canvas: bbox is " . join( ", ", @bbox ) . "\n" ;
  
  if( @bbox ) {
    $canvas->configure( -scrollregion => [ @bbox ] );
  }

} 

# open the "find" dialog for searching for nodes
sub find_dialog {
  my $self = shift ;

  my $dw   = $self->{mw}->Toplevel() ;
  $dw->attributes( -topmost => 1 ) ;
  $dw->title( 'Find' ) ;
  $dw->bind( '<Escape>' => sub { $dw->destroy() } ) ; 

  my $f = $dw->Frame()->pack() ;
  $f->Label( -text => 'Enter a keyword or a regular expression' )->pack() ;
  my $ff = $f->Frame()->pack() ;
  my $entry = $ff->Entry( ) ;
  $entry->configure( -validate => 'key', -validatecommand => sub { $self->find( @_ ) } ) ;
  $entry->bind( '<Key-Return>' => sub { $self->find( $entry->get() ) } ) ;
  $entry->pack( -side => 'left' ) ;
  $ff->Button( -text => 'Find' ,
    -command => sub { $self->find( $entry->get() ) }
    )->pack( -side => 'left' ) ;
  $ff->Button( -text => 'Cancel',
    -command => sub { $dw->destroy() }
    )->pack( -side => 'left' ) ;
  $entry->focus() ;

}



# highlight all nodes that match a certain pattern
sub find {
  my $self = shift ;
  my $what = shift ;
  #print "what=$what rest=@_\n" ;
  

  my $nmaster = $self->{nodemaster} ;

  $nmaster->node_unselect_all() ;
  return 1 unless $what ;
  my $nlist = $nmaster->node_list() ;

  foreach my $n ( @$nlist ) {

    if( 
       ( $n->{id}        and $n->{id}        =~ /$what/i ) or
       ( $n->{meta_info} and $n->{meta_info} =~ /$what/i ) or
       ( $n->{label}     and $n->{label}     =~ /$what/i ) 
       ) {

      $nmaster->node_select( $n ) ;
    }
  }

  return 1 ;
}

sub select_all {
  my $self = shift ;

  $self->{nodemaster}->node_select_all() ;

}


# ----------------------------------------------------------------------
#                                   ZOOMING
# ----------------------------------------------------------------------

# generic zoom
sub zoom {
  my $self = shift ;
  my $fac  = shift ;
  my $what = shift ;

  $self->{nodemaster}->zoom_selection( $fac, $what ) ;
}

# auto zoom. Doesn't work in all cases
sub zoom_auto {
  my $self = shift ;
  my $what = shift ;

  $self->zoom( '', $what ) ;
  $self->reconfigure_canvas() ;


}

# zoom in
sub zoom_in {
  my $self = shift ;
  print STDERR "Zoom in\n" ;

  $self->zoom( 1.3 ) ;
  $self->reconfigure_canvas() ;
}

# zoom out
sub zoom_out {
  my $self = shift ;
  print STDERR "Zoom out\n" ;
  $self->zoom( 0.8 ) ;

  $self->reconfigure_canvas() ;
}




# ----------------------------------------------------------------------
# graph operations: clustering etc.
# ----------------------------------------------------------------------

sub graph_compare {
  my $self = shift ;
  my %pars = @_ ;

  my $types = [ [ 'Dot files', [ '.dot' ] ], [ 'All files', '*' ] ] ;
  my $fname = $self->{mw}->getOpenFile( -filetypes => $types ) ;
  return unless $fname ;

  print STDERR "Reading graph from file $fname\n" ;

  my $cmpg  = gv::read( $fname ) ;
  my $graph = $self->{graph} ;

  my %seen ;
  for( my $n = gv::firstnode( $graph ) ; gv::ok( $n ) ; $n = gv::nextnode( $graph, $n ) ) {
  for( my $e = gv::firstedge( $n ) ; gv::ok( $e ) ; $e = gv::nextedge( $n, $e ) ) {
    my $hnoden = gv::nameof( gv::headof( $e ) ) ;
    my $tnoden = gv::nameof( gv::tailof( $e ) ) ;
    #print "looking for edge $tnoden -- $hnoden\n" ;

    my $key = join('-', sort( $hnoden, $tnoden ) ) ;
    next if $seen{$key} ;
    $seen{$key}++ ;

    my $tcmpn = gv::findnode( $cmpg, $tnoden ) ; next unless( $tcmpn ) ;
    my $hcmpn = gv::findnode( $cmpg, $hnoden ) ; next unless( $hcmpn ) ;

    my $cmpe = gv::findedge( $tcmpn, $hcmpn ) || gv::findedge( $hcmpn, $tcmpn ) ;
    next unless( $cmpe ) ;

    my $node1 = $self->{nodehash}{$hnoden} or die "Cannot find node $hnoden\n" ;

    my %settings = ( -width => 3, -fill => 'red' ) ;
    $node1->link_configure( $tnoden, %settings ) ;
    if( $pars{add} ) {
      gv::setv( $e, 'color', 'red' ) ;
      gv::setv( $e, 'penwidth', '3' ) ;
    }

  }}

  return unless $pars{add} ;

  # search for nodes present in the new graph (cmpg), but not in $graph
  for( my $n = gv::firstnode( $cmpg ) ; gv::ok( $n ) ; $n = gv::nextnode( $cmpg, $n ) ) {
  for( my $e = gv::firstedge( $n ) ; gv::ok( $e ) ; $e = gv::nextedge( $n, $e ) ) {
    my $hnoden = gv::nameof( gv::headof( $e ) ) ;
    my $tnoden = gv::nameof( gv::tailof( $e ) ) ;

    my $key = join('-', sort( $hnoden, $tnoden ) ) ;
    next if $seen{$key} ;
    $seen{$key}++ ;

    # nodes from the origin. graph; next if they are not present
    my $tgraphn = gv::findnode( $graph, $tnoden ) ; next unless( $tgraphn ) ;
    my $hgraphn = gv::findnode( $graph, $hnoden ) ; next unless( $hgraphn ) ;

    my $cmpe = gv::findedge( $tgraphn, $hgraphn ) || gv::findedge( $hgraphn, $tgraphn ) ;
    my $node1 = $self->{nodehash}{$hnoden} or die "Cannot find node $hnoden\n" ;
    my $node2 = $self->{nodehash}{$tnoden} or die "Cannot find node $tnoden\n" ;

    unless( $cmpe ) {
      print "cmpe not found, drawing edge $tnoden - $hnoden\n" ;
      $node1->addlink( $node2, -width => 1, -fill => 'grey' ) ;
      my $newedge = gv::edge( $graph, $hnoden, $tnoden ) ;
      gv::setv( $newedge, 'color', 'grey' ) ;
      $networkchanged = 1 ;
      next ;
    }
  }}


  print STDERR "network changed=$networkchanged\n" ;
}

# used by _partition_graph
sub _recursive_partition {
  my ( $node, $node2cluster, $clustername ) = @_ ;
  my $name = gv::nameof( $node ) ;
  #print STDERR "_recursive_partition: node=$node\n" ;

  unless( $name ) { die "Cannot get name of the node $node\n" ; }
  return if( $node2cluster->{$name} ) ;
  $node2cluster->{$name} = $clustername ;

  for( my $edge = gv::firstedge( $node ) ; gv::ok( $edge ) ; $edge = gv::nextedge( $node, $edge ) ) {
    for( my $partnode = gv::firstnode( $edge ) ; gv::ok( $partnode ) ; $partnode = gv::nextnode( $edge, $partnode ) ) {
      my $partname = gv::nameof( $partnode ) ;
      next if $partname eq $name ;
      next if $node2cluster->{$partname} ;
      _recursive_partition( $partnode, $node2cluster, $clustername ) ;
    }
  }
}

# partition graph into clusters
sub _partition_graph {
  my $self = shift ;
  my $graph = $self->{graph} ;
  unless( $graph ) {
    print STDERR "No graph defined\n" ;
    return ;
  }

  my $min_nodes = 3 ;
  #print "Partitioning graph, min_nodes=$min_nodes\n" ;
  my $node2cluster = { } ;

  for( my $node = gv::firstnode( $graph ) ; gv::ok( $node ) ; $node = gv::nextnode( $graph, $node ) ) {
    my $name = gv::nameof( $node ) ;
    #print STDERR "node $name\n" ;
    next if( $node2cluster->{$name} ) ;
    my $clustername = "cluster_$name" ;
    #print STDERR "new cluster $clustername\n" ;

    _recursive_partition( $node, $node2cluster, $clustername ) ;
  }

  my %clusters ;
  while( my ( $nodename, $clustname ) = each %$node2cluster ) {
    $clusters{$clustname}{$nodename}++ ;
  }

  #foreach my $cluster ( keys %clusters ) { my $size = scalar( keys %{ $clusters{$cluster} } ) ; }

  return ( \%clusters, $node2cluster ) ;
}

# returns all clusters
sub clusters {
  my $self = shift ;

  if( not $self->{clusters} or $networkchanged ) {
    ( $self->{clusters}, $self->{node2cluster} ) = $self->_partition_graph() ;
    $networkchanged = 0 ;
  }

  return $self->{clusters} ;
}

# returns the mapping of nodes to clusters
sub node2cluster {
  my $self = shift ;
  if( not $self->{node2cluster} or $networkchanged ) {
    ( $self->{clusters}, $self->{node2cluster} ) = $self->_partition_graph() ;
    $networkchanged = 0 ;
  }

  return $self->{node2cluster} ;
}

# returns the cluster in which the given node is
sub node_cluster {
  my $self = shift ;
  my ( $node ) = shift ;

  #print "node=$node\n" ;
  my $node_id = gv::nameof( $node ) ;

  my $mapping  = $self->node2cluster() ;
  my $clusters = $self->clusters() ;

  return $clusters->{$mapping->{$node_id}} ;
}

# remove clusters below a certain size
sub _remove_small_clusters {
  my $self = shift ;
  my $clustsize = shift || 2 ;

  my $clusters = $self->clusters( ) ;

  foreach my $c ( keys %$clusters ) {
    my $size = scalar ( keys %{ $clusters->{$c} } ) ;
    if( $size <= $clustsize ) {
      print STDERR "Removing cluster $c\n" ;

      foreach my $n ( keys %{ $clusters->{$c} } ) {
        print STDERR "Removing node $n\n" ;
        $self->{nodehash}{$n}->_delete() ;
      }
    }
  }
}

# change the indicator that the layout has changed and we need to repeat
# certain graph operations
sub network_changed { 
  $networkchanged++ ; 
}

# create a brand-new graph from selection
sub _new_graph_from_selection {
  my $self = shift ;

  my $selected = $self->{nodemaster}->node_selected_list ;
  my $graph    = $self->{graph} ;

  return unless $graph ;
  return unless @$selected ;
  print join(' ', @$selected ) . "\n" ;

  my $newg = $self->gv_graph_new() ;
  my %node_ids ;
  foreach my $n ( @$selected ) {
    my $id = $n->{id} ;
    $node_ids{$id}++ ;
  }

  #print join( ' ', keys %node_ids ) ;
  #print "\n" ;

  foreach my $n ( keys %node_ids ) {
    my $snode = gv::findnode( $graph, $n ) ;
    gv_copy_node( $graph, $newg, $snode ) ;
  }

  my %seen ;
  foreach my $n ( keys %node_ids ) {
    my $snode = gv::findnode( $graph, $n ) ;
    for( my $e = gv::firstedge( $snode ) ; gv::ok( $e ) ; $e = gv::nextedge( $snode, $e ) ) {

      my $tnoden = gv::nameof( gv::tailof( $e ) ) ;
      my $hnoden = gv::nameof( gv::headof( $e ) ) ;
      next unless( $node_ids{$tnoden} and $node_ids{$hnoden} ) ;
      next if $seen{"$hnoden-$tnoden"} ;
      next if $seen{"$hnoden-$tnoden"}++ ;

      gv_copy_edge( $graph, $newg, $e, $hnoden, $tnoden ) ;
    }

  }

  my $newmw = $self->{mw}->Toplevel() ;
  GraphEdit::GraphWindow->new( -graph => $newg, -mw => $newmw ) ;

}



#----------------------------------------------------------------------
#   FILE OPERATIONS
#----------------------------------------------------------------------

sub export {
  my $self   = shift ;
  my $format = shift || 'PDF' ;
  my $ext    = lc( $format ) ;

  my $types = [ [ "$format files", [ ".$ext" ] ], [ 'All files', '*' ] ] ;
  my $outfile = $self->{savefile} || "printout.$ext" ;
  $outfile =~ s/\.\w$/.$ext/ ;
  $outfile = $self->{mw}->getSaveFile( -filetypes => $types, -initialfile => $outfile ) ;

  return unless( $outfile ) ;
  $self->gv_render( lc($format), $outfile ) ;
}

# render the graph into specific file with a specified format
sub gv_render {
  my $self   = shift ;
  my $format = shift ;
  my $file   = shift ;

  gv::layout( $self->{graph}, 'nop2' ) ;
  gv::render( $self->{graph}, $format, $file ) ;

}

# get name of the file to save and save the graph
sub save {
  my $self   = shift ;
  my $saveas = shift || 0 ;
  my $types = [ [ 'Dot files', [ '.dot' ] ], [ 'All files', '*' ] ] ;
  my $file  = $self->{savefile} ;
  $file = '' if( $saveas ) ;
  $file ||= $self->{mw}->getSaveFile( -filetypes => $types ) ;
  return unless( $file ) ;

  $self->{savefile} = $file ;
  $self->{mw}->title( $self->{savefile} ) ;
  $self->_save_graph( $self->{savefile} ) ;
}

# save graph as a dot file.
sub _save_graph {
  my $self  = shift ;
  my $fname = shift ;
  my $graph = $self->{graph} or die "Graph not defined" ;

  gv::write( $graph, $fname ) ;
  # gv::render( $graph, 'dot', $fname ) ;
}

# preview. XXX will not work on just any system
sub preview_graph {
  my $self = shift ;
  my $tmp = 'tmpgraph.pdf' ;
  return unless( $self->{graph} ) ;

  gv::layout( $self->{graph}, 'nop2' ) ;
  gv::render( $self->{graph}, 'pdf', $tmp ) ;

  my @progs = qw/ evince acroread / ;

  foreach my $p ( @progs ) {
    unless( system "which $p" ) {
      print STDERR "Opening with $p\n" ;
      system "$p $tmp &" ;
      return ;
    }
  }

  error_box( $self->{mw}, "No suitable program for viewing found" ) ;
  print STDERR "No suitable program for viewing found\n" ;



}

sub i_read_generic {
  my $self = shift ;
  my %pars = @_ ;
  
  my $fields = $pars{fields} || [ 'auto' ] ;
  my $types  = $pars{types}  || [ [ 'Text files', [ '.txt', '.text' ] ], [ 'All files', '*' ] ] ;
  my $file = $self->{mw}->getOpenFile( -filetypes => $types ) ;
  $self->read_file( $file, @$fields ) ;
}

# field: eg. "color", "label", "meta_info"...
sub read_file {
  my $self  = shift ;
  my $ifile  = shift ;

  my @fields = @_ ;

  print STDERR "Generic file reader...\n" ;
  my $node ;

  unless( -f $ifile ) { print STDERR "File $ifile not found\n" ; return ; }
  print STDERR "Reading info from file $ifile\n" ;
  open( my $if, $ifile ) or die "Cannot read $ifile: $!\n" ;

  if( $fields[0] eq 'auto' ) {
    my $header = <$if> ;
    print "header=$header\n" ;
    chomp $header ;
    @fields = split /\s/, $header ;
    shift @fields ;
  }

  my $n = scalar @fields ;
  return unless( $n ) ;

  my $absent = 0 ;
  while( <$if> ) {

    chomp ;
    my @line = split /\s/, $_, ( $n + 1 ) ;
    my $id = shift @line ;

    unless( $node = $self->{nodehash}{$id} ) {
      #print STDERR "No such node: $id\n" ;
      $absent++ ;
      next ;
    }

    foreach my $k ( @fields ) {
      my $v = shift @line ;
      last unless $v ;
      $node->configure( $k => $v ) ;
    }

  }

  print STDERR "Done, $absent nodes in file, but not in graph\n" ;
}

# display an error message
# sample usage: error_box($mw, "Damned!") ;
sub error_box {
  my ( $mw, $msg, $fatal ) = @_;
  my $premsg;
  if ($fatal) {
    $premsg = 'Fatal error';
  } else {
    $premsg = 'Error';
  }

  my $tl = $mw->Toplevel();

  $tl->Label( -text => "$premsg:\n$msg\n" )->pack();
  $tl->grab;

  $tl->Button(
    -text    => 'OK',
    -command => sub {
      $tl->destroy;
    }
  )->pack;

  $tl->update;
  $tl->waitWindow();

  #  $tl->place(     # move it to the center of the window
  #      -x => int( $mw->width / 2 - $tl->width / 2 ),
  #      -y => int( $mw->height / 3 )
  #  );

  #my $dialog = $mw->Dialog(-text=>"$premsg:\n$msg\n",
  #-title=>$premsg,
  #-buttons=>['OK'],
  #-type=>'OK',
  #-icon=>'error'
  #) ;
  #my $answer = $dialog->Show() ;

  exit(1) if ($fatal);
}

sub view_help {
  my $self = shift ;

  my $text = <<EOF;
This is a simple graph viewer / editor. It should be used for graph files
in the DOT language format (used by the graphviz package).

Here are a few hints how to use it (everything else should be
self-explanatory):

- "meta info" does not go into dot files, but it is used when you use the
  "find" dialog.  E.g. reading some meta info (Import / Meta Info), and
  then searching (Ctrl-f) will allow you to select a category of nodes.

- if you click on a node using the *middle* button, all nodes that can be
  reached via existing edges from this node will be selected. (you select a
  cluster)

- if a selection is defined, you can alter the color / label / meta info of
  all nodes via the "properties" dialog that you can call from the context
  menu (right click on any selected node)

- zooming. ctrl-plus to zoom in, ctrl-minus to zoom out. ctrl-z zooms such
  that whatever you are zooming should afterwards fill the available screen
  (e.g. if the graph is to large, hitting ctrl-z should bring it as a whole
  to your view). If nodes are selected, though, then only the selected
  nodes will be zoomed. This allows you to selectively zoom in / zoom out
  clusters. Eg. to blow up a cluster such that it fills the whole view,
  click with the middle mouse button on one of its nodes and press ctrl-z

- if you want to set nodes color / label / meta info all at once, you can
  use the "Automatic" entry from the "import" menu. The file must have a
  header, first field should be the node id (as found in the dot file),
  next fields define what properties of the nodes will be modified (you can
  use "label", "color" and "meta_info" ). Case sensitive!

EOF

  my $helpwin = $self->{mw}->Toplevel( -title => 'Instructions' ) ;
  $helpwin->bind( '<Escape>' => sub { $helpwin->destroy() } ) ; 
  my $f = $helpwin->Frame()->pack() ;
  my $t = $f->Scrolled( 'ROText', -scrollbars => 'ow', -bg => 'white' )->pack() ;
  $t->Insert( $text ) ;
  $t->yviewMoveto( 0 ) ;
  $f->Button( -text => 'Close',
    -command => sub { $helpwin->destroy() }
    )->pack( ) ;


}


1;
