package GraphEdit::Nodemaster ;

# objects of this class hold information about a number of nodes and edges,
# allow moving them, selecting and changing their properties 
# in essence, this is an interface to Domgraphnode.pm.

use strict ;
use warnings ;
use Carp ;
use Data::Dumper ;
use Tk::Spectrum ;

use GraphEdit::GraphNode ;
use GraphEdit::Utils qw/ gv_pos_to_coord / ;
use gv ;


# class constructor
sub new {
  my $class = shift ;
  my %pars  = @_ ;

  my $canvas   = $pars{canvas}   || die "No canvas parameter provided" ;
  my $domgraph = $pars{domgraph} || die "Domgraph object not provided" ;
  my $mw       = $pars{mw}       || die "Main window object not provided" ;

  my $self = bless {
    canvas       => $canvas,
    domgraph     => $domgraph,
    mw           => $mw,
    selected     => { },
    node_hash    => { },
    edge_hash    => { },
    cv_id_to_id  => { },
    last_id      => 0,
  } ;


  $canvas->CanvasBind( '<1>' => sub { $self->canvas_clicked() } ) ;
  $canvas->bind( 'all', '<ButtonPress>',   sub { $self->{item_clicked} = 1 }  ) ;
  $canvas->bind( 'all', '<ButtonRelease>', sub { $self->{item_clicked} = 0 }  ) ;

  return $self ;
}


# ----------------------------------------------------------------------
#                          creating and managing nodes
# ----------------------------------------------------------------------

# wrapper for Domgraphnode->new
sub new_node {
  my $self = shift ;
  my %pars = @_ ;

  $pars{id} ||= ++$self->{last_id} ;

  my $canvas  = $self->{canvas} ;
  my $newnode = GraphEdit::GraphNode->new( 
          mw     => $self->{mw},
          canvas => $canvas,
          %pars,
          nodemaster => $self,
          ) ;

  $self->{node_hash}{$pars{id}} = $newnode ;
  my $cv_id = $newnode->{box_id} ;
  $self->{cv_id_to_id}{$cv_id} = $pars{id} ;
  $cv_id = $newnode->{cv_id} ;

  $canvas->bind( 'all', '<ButtonPress>',   sub { $self->{item_clicked} = 1 }  ) ;
  $canvas->bind( 'all', '<ButtonRelease>', sub { $self->{item_clicked} = 0 }  ) ;

  $canvas->bind($cv_id, '<Shift-ButtonPress-1>', sub { $self->node_toggle_select( $newnode ) } ) ;
  $canvas->bind($cv_id, '<1>', sub { $self->_move_start( $newnode ) } ) ;
  $canvas->bind($cv_id, '<3>', sub { $self->node_menu( $newnode ) }  ) ;
  $canvas->bind($cv_id, '<2>', sub { $self->cluster_select( $newnode ) }  ) ;

  return $newnode ;
}

# wrapper for "addlink" method of a node object
sub link_add {
  my $self = shift ;
  my ( $n1, $n2, %pars ) = @_ ;

  $n1->addlink( $n2, %pars ) ;

}

# return a hash ref of selected nodes
sub node_hash {
  my $self = shift ;
  return $self->{node_hash} ;
}

# return a list of nodes
sub node_list {
  my $self = shift ;
  return [ values %{ $self->{node_hash} } ] ;
}


# ----------------------------------------------------------------------
#                          node menu
# ----------------------------------------------------------------------

sub node_menu {
  my $self = shift ;
  my $node = shift ;

  my $mw   = $self->{mw} or die "mw not defined" ;
  my $menu = $self->{mw}->Menu() ;
  my $label ;
  if( @{ $self->node_selected_list } ) { $label = 'Selection' ; } 
  else { $label = "Node $node->{label}" ; }

  $menu->add( 'command', -label => $label ) ;
  $menu->add( 'separator' ) ;
  $menu->Popup( -popover => 'cursor' ) ;
  $menu->add( 'command', -label => 'Delete node',   -command => sub { $node->_delete() ; } ) ;
  $menu->add( 'command', -label => 'Copy label',    -command => sub { $self->clipboardCopy( $node ) } ) ;
  $menu->add( 'command', -label => 'Node color',    -command => sub { $self->node_change_color( $node ) } ) ;
  $menu->add( 'command', -label => 'Properties...', -command => sub { $self->node_conf_gui( $node ) } ) ;
  $self->{item_clicked} = 0 ;
}

sub clipboardCopy {
  my $self = shift ;
  my $node = shift ;

  my $selected = $self->node_selected_list() ;
  unless( @$selected ) {
    $selected = [ $node ] ;
  }

  my $text = '' ;
  foreach my $node ( @$selected ) {
    $text .= $node->{label} || $node->{id} ;
    $text .= "\n" ;
  }

  my $canvas = $self->{canvas} ;
  $canvas->clipboardClear() ;
  $canvas->clipboardAppend( $text ) ;
}

# dialog for changing node color
sub node_change_color {
  my $self = shift ;
  my $node = shift ;

  my $selected = $self->node_selected_list ;
  unless( @$selected ) { $selected = [ $node ] ; }

  my $color = 
      $self->{mw}->Spectrum( 
      -title        => 'Pick a Color',
      -buttons      => [ 'Ok', 'Cancel' ], ) ;
      #-initialcolor => $old_color ) ;


  foreach my $n ( @$selected ) {
    $n->configure( color => $color ) ;
  }

}

# the dialog for node properties
sub node_conf_gui {
  my $self = shift ;
  my $node = shift ;
  my $dw   = $self->{mw}->Toplevel() ;
  my ( $f, $ff, $entry ) ;
  my $conf = { } ;

  $dw->bind( '<Key-Return>' => sub { $self->node_conf_apply( $conf, $node ) } ) ;
  $dw->attributes( -topmost => 1 ) ;
  $dw->title( 'Node properties' ) ;
  $dw->bind( '<Escape>' => sub { $dw->destroy() } ) ; 
  $f = $dw->Frame()->pack() ;

  my $selected = $self->node_selected_list() ;
  $ff = $f->Frame()->pack() ;
  if( @$selected ) {
    my $n = scalar @$selected ;
    $ff->Label( -text => "Properties of $n selected nodes" )->pack( -side => 'left' ) ;
  } else {
    $selected = [ $node ] ;
    $ff->Label( -text => "Properties of node $node->{id}" )->pack( -side => 'left' ) ;
    $conf->{label}     = $node->{label} ;
    $conf->{color}     = $node->{color} ;
    $conf->{meta_info} = $node->{meta_info} ;
  }


  $ff = $f->Frame()->pack( -fill => 'x' ) ;
  $ff->Label( -text => 'Node Label:' )->pack( -side => 'left'  ) ;
  $entry = $ff->Entry( -textvar => \$conf->{label} || '' ) ;
  $entry->pack( -side => 'left' ) ;
  

  $ff = $f->Frame()->pack( -fill => 'x' ) ;
  $ff->Label( -text => 'Color: ' )->pack( -side => 'left' ) ;
  $self->color_sel_button( $ff, \$conf->{color} ) ;

  $ff = $f->Frame()->pack(-fill => 'x') ;
  $ff->Label( -text => "Meta information" )->pack( -side => 'left' ) ;
  $ff = $f->Frame()->pack() ;
  $conf->{meta_info} = $ff->Text( -height => 4, -width => 30, ) ;
  $conf->{meta_info}->pack() ;


  #$entry->configure( -validate => 'key', -validatecommand => sub { $self->find( @_ ) } ) ;
  #$entry->bind( '<Key-Return>' => sub { $self->find( $entry->get() ) } ) ;

  $ff = $f->Frame()->pack() ;

  $ff->Button( -text => 'Cancel',
    -command => sub { $dw->destroy() }
    )->pack( -side => 'left' ) ;
  $ff->Button( -text => 'Apply' ,
    -command => sub { $self->node_conf_apply( $conf, $node ) }
    )->pack( -side => 'left' ) ;
  $ff->Button( -text => 'OK' ,
    -command => sub { 
      $self->node_conf_apply( $conf, $node ) ;
      $dw->destroy() ;
       }
    )->pack( -side => 'left' ) ;


}

# the color select button
sub color_sel_button {
  my $self = shift ;
  my ( $ff, $varref ) = @_ ;

  my $but ;
  my $background = $$varref || 'red' ;
  
  $but = $ff->Button( 
    -height  => 0.5,
    -width   => 5,
    -command => sub {
      my $color = 
      $self->{mw}->Spectrum( 
      -title        => 'Pick a Color',
      -buttons      => [ 'Ok', 'Cancel' ], ) ;
      return unless $color ;
      $but->configure( -background => $color ) ;
      $$varref = $color ;
  }, -background => $background ) ;

  $but->pack() ;

}

# apply configuration to a node or selection
sub node_conf_apply {
  my $self = shift ;
  my ( $conf, $node ) = @_ ;

  $conf->{meta_info} = $conf->{meta_info}->Contents() ;

  my $selected = $self->node_selected_list() ;
  unless( @$selected ) { $selected = [ $node ] ; }

  foreach my $n ( @$selected ) {
    $n->configure( %$conf ) ;
  }

}


# ----------------------------------------------------------------------
#                          node selecting
# ----------------------------------------------------------------------

sub node_select_all {
  my $self = shift ;

  my $allnodes = $self->node_list ;
  foreach my $node ( @$allnodes ) {
    $self->node_select( $node ) ;
  }

}

sub node_selected_list {
  my $self = shift ;

  return [ values %{ $self->{selected} } ] ;
}

sub node_select {
  my $self = shift ;
  my $node = shift ;
  return if( $node->{mark} ) ;

  my $id   = $node->{id} ;
  $node->mark() ;

  $self->{selected}{$id} = $node ;
}

sub node_toggle_select {
  my $self = shift ;
  my $node = shift ;
  my $id   = $node->{id} ;

  if( $self->{selected}{$id} ) {
    $self->node_unselect( $node ) ;
  } else {
    $self->node_select( $node ) ;
  }

}

sub node_unselect {
  my $self = shift ;
  my $node = shift ;
  my $id   = $node->{id} ;

  return unless( $self->{selected}{$id} ) ;

  $node->unmark() ;
  delete $self->{selected}{$id} ;

}

sub node_unselect_all {
  my $self = shift ;

  my $nodelist = $self->node_list() ;

  foreach my $node ( @$nodelist ) { $self->node_unselect( $node ) ; }
}




# ----------------------------------------------------------------------
#           moving a selection of nodes
# ----------------------------------------------------------------------

sub _move {
  my $self = shift ;
  my ( $selected, $id ) = @_ ;
  my $canvas = $self->{canvas} ;

  my $curx = $canvas->canvasx($Tk::event->x) ;
  my $cury = $canvas->canvasy($Tk::event->y) ;

  my $dx = $curx - $self->{prevx} ;
  my $dy = $cury - $self->{prevy} ;
  ( $self->{prevx}, $self->{prevy} ) = ( $curx, $cury ) ;

  foreach my $node ( @$selected ) {
    $node->_move_graphics_rel( $dx, $dy ) ;
  }
}

sub _move_end {
  my $self = shift ;
  my ( $selected, $id ) = @_ ;

  my $canvas = $self->{canvas} ;

  foreach my $node ( @$selected ) {
    if( $node->{gvnode} ) {
      $node->_move_gv_rel( $self->{startx} - $self->{prevx}, $self->{starty} - $self->{prevy} ) 
    }
  }

  #my $id = $node->{cv_id} ;
  $canvas->bind( $id, '<Motion>', '') ;
}

sub _move_start {
  my $self = shift ;
  my $node = shift ;

  my $canvas   = $self->{canvas} ;
  my $selected = $self->node_selected_list() ;

  unless( $node->{mark} ) {
    $self->node_unselect_all() ;
    $selected = [ $node ] ;
  }

  my $id = $node->{cv_id} ;

  $self->{prevx} = $canvas->canvasx($Tk::event->x) ;
  $self->{prevy} = $canvas->canvasy($Tk::event->y) ;
  ( $self->{startx}, $self->{starty} ) = ( $self->{prevx}, $self->{prevy} ) ;

  $canvas->bind( $id, '<Motion>', sub { $self->_move( $selected, $id ) } ) ;
  $canvas->bind( $id, '<ButtonRelease-1>', sub { $self->_move_end( $selected, $id ) } ) ;
  $canvas->bind( $id, '<Leave>', '' ) ; 

}









# ----------------------------------------------------------------------
#           this is to select a whole cluster
# ----------------------------------------------------------------------

sub cluster_select {
  my $self = shift ;
  my $node = shift ;

  my $domgraph = $self->{domgraph} ;

  $self->node_unselect_all() ;
  my $gvnode = $node->{gvnode} ;

  return unless( $gvnode ) ;

  my $cluster = $domgraph->node_cluster( $gvnode ) ;

  foreach my $id ( keys %$cluster ) {
    my $node = $self->{node_hash}->{$id} ;
    $self->node_select( $node ) ;
  }



}

# zoom in / zoom out a selection of nodes
sub zoom_selection {
  my $self     = shift ;
  my $fac      = shift ;
  my $selected = $self->node_selected_list() ;
  my $canvas   = $self->{canvas} ;

  unless( @$selected ) {
    $selected = $self->node_list() ;
  }

  my ( $av_x, $av_y, $n ) ;
  my ( $minx, $miny, $maxx, $maxy ) ;

  foreach my $node ( @$selected ) {
    my ( $x, $y ) = gv_pos_to_coord( gv::getv( $node->{gvnode}, 'pos' ) ) ;
    $y = -$y ;
    $av_x += $x ; $av_y += $y ;
    $minx = $x if( not defined $minx or $x < $minx ) ;
    $miny = $y if( not defined $miny or $y < $miny ) ;
    $maxx = $x if( not defined $maxx or $x > $maxx ) ;
    $maxy = $y if( not defined $maxy or $y > $maxy ) ;

    $n++ ;
  }

  $av_x /= $n ;
  $av_y /= $n ;

  # if factor undefined, we want to derive it automatically somehow
  my ( $dx, $dy ) ;
  unless( $fac ) {
    print STDERR "autoscaling\n" ;
    my $viewport = viewport( $canvas ) ;

    $av_x = ( $maxx + $minx ) / 2 ;
    $av_y = ( $maxy + $miny ) / 2 ;

    $dx = ( $viewport->[0] + $viewport->[2] ) / 2 - $av_x ;
    $dy = ( $viewport->[1] + $viewport->[3] ) / 2 - $av_y ;
    my $vposx = ( $viewport->[0] + $viewport->[2] ) / 2 ;
    my $vposy = ( $viewport->[1] + $viewport->[3] ) / 2 ;

    my $fac1 = 0.9 * ( $viewport->[2] - $viewport->[0] ) / ( $maxx - $minx ) ;
    my $fac2 = 0.9 * ( $viewport->[3] - $viewport->[1] ) / ( $maxy - $miny ) ;
    $fac = $fac1 < $fac2 ? $fac1 : $fac2 ;

    foreach my $node ( @$selected ) {
      $node->move( $dx, $dy ) ,
    }

    $av_x = $vposx ;
    $av_y = $vposy ;


    #return ;
  }

  foreach my $node ( @$selected ) {
    $node->scale_pos( $fac, $av_x, $av_y ) ;
  }

}

sub viewport {
  my $canvas = shift ;

  my $sr = $canvas->cget( '-scrollregion' ) ;
  my $xv = $canvas->xview() ;
  my $yv = $canvas->yview() ;

  my @bb = (
    ( $sr->[2] - $sr->[0]) * $xv->[0] + $sr->[0],
    ( $sr->[3] - $sr->[1]) * $yv->[0] + $sr->[1],
    ( $sr->[2] - $sr->[0]) * $xv->[1] + $sr->[0],
    ( $sr->[3] - $sr->[1]) * $yv->[1] + $sr->[1],
    ) ;

  return \@bb ;
}

# ----------------------------------------------------------------------
#           this is to select different nodes with a mask
# ----------------------------------------------------------------------

sub canvas_clicked {
  my $self = shift ;

  return if $self->{item_clicked} ;
  $self->node_unselect_all() ;
  my $canvas = $self->{canvas} ;

  #return if $item_clicked ;

  $self->{startx} = $canvas->canvasx($Tk::event->x) ;
  $self->{starty} = $canvas->canvasy($Tk::event->y) ;
  ( $self->{prevx}, $self->{prevy} ) = ( $self->{startx}, $self->{starty} ) ;

  my @bb = ( $self->{startx}, $self->{starty}, 
             $self->{startx}, $self->{prevy}, 
             $self->{prevx},  $self->{prevy}, 
             $self->{prevx},  $self->{starty} ) ;

  my $poly = $canvas->createPolygon(
    @bb,
    -fill    => '',
    -outline => 'black',
  );

  $canvas->CanvasBind( '<Motion>',          sub { $self->_move_selection_mask( $poly ) } ) ;
  $canvas->CanvasBind( '<ButtonRelease-1>', sub { $self->_selection_mask_end( $poly ) } ) ;
  #$canvas->bind( $id, '<Leave>', '' ) ; 


}

sub _selection_mask_end {
  my $self = shift ;
  my $id   = shift ;
  my $canvas = $self->{canvas} ;
  $canvas->CanvasBind( '<Motion>', '' ) ;
  $canvas->delete( $id ) ;
}

sub _move_selection_mask {
  my $self = shift ;
  my $id   = shift ;

  my $canvas = $self->{canvas} ;

  $self->{prevx} = $canvas->canvasx($Tk::event->x) ;
  $self->{prevy} = $canvas->canvasy($Tk::event->y) ;

  my @bb = ( $self->{startx}, $self->{starty}, 
             $self->{startx}, $self->{prevy}, 
             $self->{prevx},  $self->{prevy}, 
             $self->{prevx},  $self->{starty} ) ;

  $canvas->coords( $id, @bb ) ;
  @bb = $canvas->bbox( $id ) ;

  my $list = $canvas->find( 'enclosed', @bb ) ;

  my %new ;
  foreach my $l ( @$list ) {
    my $node_id = $self->{cv_id_to_id}{ $l } ;
    next unless $node_id ;
    $new{ $node_id }++ ;
  }

  while( my ( $node_id, $node ) = each %{ $self->node_hash } ) {
    unless( $new{$node_id} ) { $self->node_unselect( $node ) }
  }

  foreach my $node_id ( keys %new ) {
    my $node = $self->{node_hash}{$node_id} ;
    $self->node_select( $node ) ;
  }


}



1;
